import 'package:bloc_stream/bloc_stream.dart';
import 'package:riverpod/riverpod.dart';
import 'package:riverpod/src/value_provider.dart';
import 'package:meta/meta.dart';

// ignore: implementation_imports
import 'package:riverpod/src/framework.dart';

import 'builders.dart';

abstract class BlocStreamProviderRef<Bloc extends BlocStream<Value>, Value>
    implements Ref {
  /// The [BlocStream] currently exposed by this provider.
  ///
  /// Cannot be accessed while creating the provider.
  BlocStream get bloc;
}

// ignore: subtype_of_sealed_class
mixin BlocStreamProviderOverrideMixin<Bloc extends BlocStream<Value>, Value>
    on ProviderBase<Value> {
  ProviderBase<Bloc> get bloc;

  @override
  late final List<ProviderOrFamily>? dependencies = [bloc];

  @override
  ProviderBase<Bloc> get originProvider => bloc;

  Override overrideWithValue(Bloc value) {
    return ProviderOverride(
      origin: bloc,
      override: ValueProvider<Bloc>(value),
    );
  }
}

// ignore: subtype_of_sealed_class
@sealed
class BlocStreamProvider<Bloc extends BlocStream<Value>, Value>
    extends AlwaysAliveProviderBase<Value>
    with
        BlocStreamProviderOverrideMixin<Bloc, Value>,
        OverrideWithProviderMixin<Bloc, BlocStreamProvider<Bloc, Value>> {
  BlocStreamProvider(
    Create<Bloc, BlocStreamProviderRef<Bloc, Value>> create, {
    String? name,
    List<ProviderOrFamily>? dependencies,
    Family? from,
    Object? argument,
  })  : bloc = _BlocStreamProvider(
          create,
          name: name,
          dependencies: dependencies,
          from: from,
          argument: argument,
        ),
        super(name: name, from: from, argument: argument);

  static const family = BlocStreamProviderFamilyBuilder();
  static const autoDispose = AutoDisposeBlocStreamProviderBuilder();

  @override
  late final AlwaysAliveProviderBase<Bloc> bloc;

  @override
  Value create(ProviderElementBase<Value> ref) {
    final bloc = ref.watch(this.bloc);

    ref.onDispose(bloc.listen((newValue) {
      ref.setState(newValue);
    }).cancel);

    return bloc.value;
  }

  @override
  bool updateShouldNotify(Value previousState, Value newState) {
    return true;
  }

  @override
  ProviderElementBase<Value> createElement() => ProviderElement(this);
}

// ignore: subtype_of_sealed_class
class _BlocStreamProvider<Bloc extends BlocStream<Value>, Value>
    extends AlwaysAliveProviderBase<Bloc> {
  _BlocStreamProvider(
    this._create, {
    required String? name,
    required this.dependencies,
    Family? from,
    Object? argument,
  }) : super(
          name: name == null ? null : '$name.bloc',
          from: from,
          argument: argument,
        );

  final Create<Bloc, BlocStreamProviderRef<Bloc, Value>> _create;

  @override
  final List<ProviderOrFamily>? dependencies;

  @override
  Bloc create(
    covariant BlocStreamProviderRef<Bloc, Value> ref,
  ) {
    final bloc = _create(ref);
    ref.onDispose(bloc.close);
    return bloc;
  }

  @override
  bool updateShouldNotify(Bloc previousState, Bloc newState) {
    return true;
  }

  @override
  _BlocStreamProviderElement<Bloc, Value> createElement() =>
      _BlocStreamProviderElement(this);
}

// ignore: subtype_of_sealed_class
@sealed
class _BlocStreamProviderElement<Bloc extends BlocStream<Value>, Value>
    extends ProviderElementBase<Bloc>
    implements BlocStreamProviderRef<Bloc, Value> {
  _BlocStreamProviderElement(
    _BlocStreamProvider<Bloc, Value> provider,
  ) : super(provider);

  @override
  Bloc get bloc => requireState;
}

// ignore: subtype_of_sealed_class
@sealed
class BlocStreamProviderFamily<Bloc extends BlocStream<Value>, Value, Arg>
    extends Family<Value, Arg, BlocStreamProvider<Bloc, Value>> {
  BlocStreamProviderFamily(
    this._create, {
    String? name,
    List<ProviderOrFamily>? dependencies,
  }) : super(
          name: name,
          dependencies: dependencies,
        );

  final FamilyCreate<Bloc, BlocStreamProviderRef<Bloc, Value>, Arg> _create;

  @override
  BlocStreamProvider<Bloc, Value> create(
    Arg argument,
  ) =>
      BlocStreamProvider<Bloc, Value>(
        (ref) => _create(ref, argument),
        name: name,
        from: this,
        argument: argument,
      );

  @override
  void setupOverride(Arg argument, SetupOverride setup) {
    final provider = call(argument);
    setup(origin: provider.bloc, override: provider.bloc);
  }

  Override overrideWithProvider(
    BlocStreamProvider<Bloc, Value> Function(Arg argument) override,
  ) =>
      FamilyOverride<Arg>(this, (arg, setup) {
        final provider = call(arg);
        setup(origin: provider.bloc, override: override(arg).bloc);
      });
}

/// {@macro riverpod.providerrefbase}
abstract class AutoDisposeBlocStreamProviderRef<Bloc extends BlocStream<Value>,
    Value> implements AutoDisposeRef {
  /// The [BlocStream] currently exposed by this provider.
  ///
  /// Cannot be accessed while creating the provider.
  Bloc get bloc;
}

// ignore: subtype_of_sealed_class
class _AutoDisposeBlocStreamProvider<Bloc extends BlocStream<Value>, Value>
    extends AutoDisposeProviderBase<Bloc> {
  _AutoDisposeBlocStreamProvider(
    this._create, {
    required String? name,
    required this.dependencies,
    Family? from,
    Object? argument,
  }) : super(
          name: name == null ? null : '$name.bloc',
          from: from,
          argument: argument,
        );

  final Create<Bloc, AutoDisposeBlocStreamProviderRef<Bloc, Value>> _create;

  @override
  final List<ProviderOrFamily>? dependencies;

  @override
  Bloc create(covariant AutoDisposeBlocStreamProviderRef<Bloc, Value> ref) {
    final bloc = _create(ref);
    ref.onDispose(bloc.close);
    return bloc;
  }

  @override
  bool updateShouldNotify(Bloc previousValue, Bloc newValue) {
    return true;
  }

  @override
  _AutoDisposeBlocStreamProviderElement<Bloc, Value> createElement() {
    return _AutoDisposeBlocStreamProviderElement(this);
  }
}

class _AutoDisposeBlocStreamProviderElement<Bloc extends BlocStream<Value>,
        Value> extends AutoDisposeProviderElementBase<Bloc>
    implements AutoDisposeBlocStreamProviderRef<Bloc, Value> {
  _AutoDisposeBlocStreamProviderElement(
    _AutoDisposeBlocStreamProvider<Bloc, Value> provider,
  ) : super(provider);

  @override
  Bloc get bloc => requireState;
}

// ignore: subtype_of_sealed_class
/// {@macro riverpod.blocstreamprovider}
class AutoDisposeBlocStreamProvider<Bloc extends BlocStream<Value>, Value>
    extends AutoDisposeProviderBase<Value>
    with
        BlocStreamProviderOverrideMixin<Bloc, Value>,
        OverrideWithProviderMixin<Bloc,
            AutoDisposeBlocStreamProvider<Bloc, Value>> {
  /// {@macro riverpod.blocstreamprovider}
  AutoDisposeBlocStreamProvider(
    Create<Bloc, AutoDisposeBlocStreamProviderRef<Bloc, Value>> create, {
    String? name,
    List<ProviderOrFamily>? dependencies,
    Family? from,
    Object? argument,
  })  : bloc = _AutoDisposeBlocStreamProvider(
          create,
          name: name,
          dependencies: dependencies,
          from: from,
          argument: argument,
        ),
        super(name: name, from: from, argument: argument);

  /// {@macro riverpod.family}
  static const family = AutoDisposeBlocStreamProviderFamilyBuilder();

  /// {@template riverpod.blocstreamprovider.bloc}
  /// Obtains the [BlocStream] associated with this [AutoDisposeBlocStreamProvider],
  /// without listening to it.
  ///
  /// Listening to this provider may cause providers/widgets to rebuild in the
  /// event that the [BlocStream] it recreated.
  /// {@endtemplate}
  @override
  final AutoDisposeProviderBase<Bloc> bloc;

  @override
  Value create(AutoDisposeProviderElementBase<Value> ref) {
    final bloc = ref.watch(this.bloc);

    ref.onDispose(bloc.listen((newValue) {
      ref.setState(newValue);
    }).cancel);

    return bloc.value;
  }

  @override
  bool updateShouldNotify(Value previousValue, Value newValue) {
    return true;
  }

  @override
  AutoDisposeProviderElementBase<Value> createElement() {
    return AutoDisposeProviderElement(this);
  }
}

// ignore: subtype_of_sealed_class
/// {@template riverpod.blocstreamprovider.family}
/// A class that allows building a [AutoDisposeBlocStreamProvider] from an external parameter.
/// {@endtemplate}
@sealed
class AutoDisposeBlocStreamProviderFamily<Bloc extends BlocStream<Value>, Value,
        Arg>
    extends Family<Value, Arg, AutoDisposeBlocStreamProvider<Bloc, Value>> {
  /// {@macro riverpod.blocstreamprovider.family}
  AutoDisposeBlocStreamProviderFamily(
    this._create, {
    String? name,
    List<ProviderOrFamily>? dependencies,
  }) : super(name: name, dependencies: dependencies);

  final FamilyCreate<Bloc, AutoDisposeBlocStreamProviderRef<Bloc, Value>, Arg>
      _create;

  @override
  AutoDisposeBlocStreamProvider<Bloc, Value> create(Arg argument) =>
      AutoDisposeBlocStreamProvider<Bloc, Value>(
        (ref) => _create(ref, argument),
        name: name,
        from: this,
        argument: argument,
      );

  @override
  void setupOverride(Arg argument, SetupOverride setup) {
    final provider = call(argument);
    setup(origin: provider, override: provider);
    setup(origin: provider.bloc, override: provider.bloc);
  }

  /// {@macro riverpod.overridewithprovider}
  Override overrideWithProvider(
    AutoDisposeBlocStreamProvider<Bloc, Value> Function(Arg argument) override,
  ) =>
      FamilyOverride<Arg>(
        this,
        (arg, setup) {
          final provider = call(arg);
          setup(origin: provider.bloc, override: override(arg).bloc);
        },
      );
}
